from collections import defaultdict

import numpy as np
import os.path

from pyTMHMM.model import parse
from pyTMHMM.hmm import viterbi, forward, backward


GROUP_NAMES = ('i', 'm', 'o')

DEFAULT_MODEL = os.path.join(os.path.dirname(__file__), 'TMHMM2.0.model')

def predict(sequence, model_or_filelike=DEFAULT_MODEL, compute_posterior=True):
    if isinstance(model_or_filelike, tuple):
        model = model_or_filelike
    else:
        header, model = parse(model_or_filelike)

    _, path = viterbi(sequence, *model)

    if compute_posterior:
        forward_table, constants = forward(sequence, *model)
        backward_table = backward(sequence, constants, *model)

        posterior = forward_table * backward_table
        _, _, _, char_map, label_map, name_map = model

        observations = len(sequence)
        states = len(name_map)

        table = np.zeros(shape=(observations, 3))
        for i in range(observations):
            group_probs = defaultdict(float)
            for j in range(states):
                group = label_map[j].lower()
                group_probs[group] += posterior[i, j]

            for k, group in enumerate(GROUP_NAMES):
                table[i, k] = group_probs[group]
        return path, table/table.sum(axis=1, keepdims=True)
    return path



import collections
import re
import os

import numpy as np


def _tokenize(contents):
    return re.findall(r'([A-Za-z0-9\.\-_]+|[:;\{\}])', contents)


def _strip_comments(file_like):
    with open(file_like, 'r') as f:
        lines = f.readlines()
    return ''.join(filter(lambda l: not l.startswith('#'), lines))


def _parse_list(tokens):
    parsed_list = []
    while True:
        token = tokens.popleft()
        if token == ';':
            tokens.appendleft(token)
            return tokens, parsed_list
        parsed_list.append(token)


def _parse_map(tokens):
    parsed_map = collections.OrderedDict()
    while True:
        token = tokens.popleft()
        if token == ';':
            tokens.appendleft(token)
            return tokens, parsed_map
        next_token = tokens.popleft()

        # Fallback if the map was actually a list
        if next_token != ':':
            tokens.appendleft(next_token)
            tokens.appendleft(token)
            return tokens, None

        value = tokens.popleft()
        parsed_map[token] = float(value)



def _parse_state(tokens):
    state_name = tokens.popleft()
    tokens.popleft() # "{"

    parsed_state = {}
    while True:
        token = tokens.popleft()
        if token == '}':
            return tokens, (state_name, parsed_state)
        if token in ('trans', 'only'):
            tokens, value = _parse_map(tokens)
            if value is None:
                tokens, value = _parse_list(tokens)
        elif token in ('type', 'end'):
            value = int(tokens.popleft())
        else:
            value = tokens.popleft()
        parsed_state[token] = value
        tokens.popleft() # ";"


def _parse_header(tokens):
    tokens.popleft() # "header"
    tokens.popleft() # "{"

    header = {}
    while True:
        token = tokens.popleft()
        if token == '}':
            break
        header[token] = tokens.popleft()
        tokens.popleft() # ";"
    return tokens, header


def _normalize_states(states):
    """Normalize states by inheriting parameters explicitly.

    The TMHMM file format allows parameters to be tied to the parameters of
    some other state. This basically means that a state inherits the parameters
    from another state.

    The normalization performed by this function consists of copying the
    specified parameters from the parent state to the inheriting state such
    that all states explicitly specify their transition and emission
    probabilities.
    """
    for name, state in states.items():
        # inherit parent's transition probabilities, but only for
        # the states specified for this state.
        if 'tied_trans' in state:
            parent_state = states[state['tied_trans']]
            to_states = state['trans']
            states[name]['trans'] = dict(zip(state['trans'],
                                         parent_state['trans'].values()))

        # inherit parent's emission probabilities
        if 'tied_letter' in state:
            parent_state = state['tied_letter']
            states[name]['only'] = dict(states[parent_state]['only'])
    return states


def _to_matrix_form(alphabet, states):
    """
    Convert a model to matrix form.
    """
    # pull out initial probabilities
    begin = dict(states['begin'])
    del states['begin']

    # build state -> index mapping
    state_map = {v: k for k, v in enumerate(states)}
    # build character -> index mapping
    char_map = {v: k for k, v in enumerate(alphabet)}

    no_states = len(states)

    initial = np.zeros(shape=(no_states,))
    transitions = np.zeros(shape=(no_states, no_states))
    emissions = np.zeros(shape=(no_states, len(alphabet)))

    label_map = {}
    name_map = dict(enumerate(states))

    # initial probabilities
    for state_name, trans_prob in begin['trans'].items():
        this_state_idx = state_map[state_name]
        initial[this_state_idx] = trans_prob

    for state_name, state in states.items():
        this_state_idx = state_map[state_name]

        # label map
        if 'label' in state:
            label_map[this_state_idx] = state['label']

        # transition probabilities
        for other_state_name, trans_prob in state['trans'].items():
            other_state_idx = state_map[other_state_name]
            transitions[this_state_idx, other_state_idx] = trans_prob

        # emission probabilities
        for character, emission_prob in state['only'].items():
            this_character_idx = char_map[character]
            emissions[this_state_idx, this_character_idx] = emission_prob

    return initial, transitions, emissions, char_map, label_map, name_map


def parse(file_like):
    """
    Parse a model in the TMHMM 2.0 format.

    :param file_like: a file-like object to read and parse.
    :return: a model
    """
    contents = _strip_comments(file_like)
    tokens = collections.deque(_tokenize(contents))

    tokens, header = _parse_header(tokens)

    states = {}
    while tokens:
        tokens, (name, state) = _parse_state(tokens)
        states[name] = state

    assert not tokens, "list of tokens not consumed completely"
    return header, _to_matrix_form(header['alphabet'],
                                   _normalize_states(states))




import numpy as np

cimport numpy as np
cimport cython

DTYPE = np.double
ctypedef np.double_t DTYPE_t


@cython.boundscheck(False)
def viterbi(sequence,
            np.ndarray[DTYPE_t, ndim=1] initial,
            np.ndarray[DTYPE_t, ndim=2] transitions,
            np.ndarray[DTYPE_t, ndim=2] emissions,
            char_map, label_map, name_map):
    """
    Compute the most probable path through the model given the sequence.

    This function implements Viterbi's algorithm in log-space.

    :param sequence str: a string over the alphabet specified by the model.
    :rtype: tuple(matrix, optimal_path)
    :return: a tuple consisting of the dynamic programming table and the
             optimal path.
    """
    sequence = sequence.upper()

    cdef int no_observations = len(sequence)
    cdef int no_states = len(initial)

    cdef float neginf = -np.inf

    # work in log space
    initial = np.log(initial)
    transitions = np.log(transitions)
    emissions = np.log(emissions)

    cdef np.ndarray[DTYPE_t, ndim=2] M = \
        np.zeros([2, no_states],dtype=DTYPE)
    cdef np.ndarray[np.int_t, ndim=2] P = \
        np.zeros([no_observations, no_states], dtype=np.int)

    cdef unsigned int i, j, k, max_state, next_state, observation
    cdef double max_state_prob, prob

    observation = char_map[sequence[0]]
    for i in range(no_states):
        M[0, i] = initial[i] + emissions[i, observation]

    for i in range(1, no_observations):
        observation = char_map[sequence[i]]
        for j in range(no_states):
            max_state = 0
            max_state_prob = neginf
            for k in range(no_states):
                prob = M[(i - 1) % 2, k] + transitions[k, j]
                if prob > max_state_prob:
                    max_state, max_state_prob = k, prob
            M[i % 2, j] = max_state_prob + emissions[j, observation]
            P[i, j] = max_state

    # TODO: figure out why stuff doesn't work when using cython without turning
    #       the range generator into a list first.
    # TODO: stuff crashes if one uses reversed(range(no_observations)), why?

    backtracked = []
    next_state = np.argmax(M[no_observations % 2,], axis=0)
    for i in list(range(no_observations - 1, -1, -1)):
        backtracked.append(label_map[next_state])
        next_state = P[i, next_state]

    return M, ''.join(reversed(backtracked))


@cython.boundscheck(False)
def forward(sequence,
            np.ndarray[DTYPE_t, ndim=1] initial,
            np.ndarray[DTYPE_t, ndim=2] transitions,
            np.ndarray[DTYPE_t, ndim=2] emissions,
            char_map, label_map, name_map):
    """
    Compute the probability distribution of states after observing the sequence.

    This function implements the scaled Forward algorithm.

    :param sequence str: a string over the alphabet specified by the model.
    :rtype: tuple(matrix, constants)
    :return: the scaled dynamic programming table and the constants used to
             normalize it.
    """

    sequence = sequence.upper()

    cdef int no_observations = len(sequence)
    cdef int no_states = len(initial)

    cdef np.ndarray[DTYPE_t, ndim=2] M = \
        np.zeros([no_observations, no_states], dtype=DTYPE)
    cdef np.ndarray[DTYPE_t, ndim=1] constants = \
        np.zeros(no_observations, dtype=DTYPE)

    cdef unsigned int i, j, k, observation
    cdef double prob, state_sum

    observation = char_map[sequence[0]]
    for i in range(no_states):
        M[0, i] = initial[i] * emissions[i, observation]
    constants[0] = np.sum(M[0])
    M[0] = M[0] / constants[0]

    for i in range(1, no_observations):
        observation = char_map[sequence[i]]
        for j in range(no_states):
            state_sum = 0.0
            for k in range(no_states):
                state_sum += M[(i - 1), k] * transitions[k, j]
            M[i, j] = state_sum * emissions[j, observation]
        constants[i] = np.sum(M[i])
        M[i] = M[i] / constants[i]

    return M, constants


@cython.boundscheck(False)
def backward(sequence,
             constants,
             np.ndarray[DTYPE_t, ndim=1] initial,
             np.ndarray[DTYPE_t, ndim=2] transitions,
             np.ndarray[DTYPE_t, ndim=2] emissions,
             char_map, label_map, name_map):
    """
    Compute the probability of being in some state and generating the rest of
    the sequence.

    This function implements the scaled backward algorithm.

    :param sequence str: a string over the alphabet specified by the model.
    :param constants np.ndarray: an array of the constants used to normalize
                                 the forward table.
    :rtype: np.ndarray
    :return: the scaled backward table.
    """

    sequence = sequence.upper()

    cdef int no_observations = len(sequence)
    cdef int no_states = len(initial)

    cdef np.ndarray[DTYPE_t, ndim=2] M = \
        np.zeros([no_observations, no_states], dtype=DTYPE)

    cdef unsigned int i, j, k, observation
    cdef double prob, state_sum

    M[no_observations - 1] = 1.0 / constants[no_observations - 1]

    for i in range(no_observations-2, -1, -1):
        observation = char_map[sequence[i]]
        for j in range(no_states):
            state_sum = 0.0
            for k in range(no_states):
                state_sum += M[(i + 1), k] * transitions[j, k]
            M[i, j] = state_sum * emissions[j, observation]
        M[i] = M[i] / constants[i]

    return M



import hashlib
import subprocess
import unittest
import sys
import os
import re
testdir = os.path.dirname(__file__)
srcdir = '../pyTMHMM'
sys.path.insert(0, os.path.abspath(os.path.join(testdir, srcdir)))
import pyTMHMM


class TestTMHMM(unittest.TestCase):
    '''
    Run tmhmm and check the *annotation file and results
    '''

    annotation = 'B9DFX7|1B|HMA8_ARATH.annotation'
    plot = 'B9DFX7|1B|HMA8_ARATH.plot'
    summary = 'B9DFX7|1B|HMA8_ARATH.summary'
    seq = 'MASNLLRFPLPPPSSLHIRPSKFLVNRCFPRLRRSRIRRHCSRPFFLVSNSVEISTQSFESTESSIESVKSITSDTPIL\
LDVSGMMCGGCVARVKSVLMSDDRVASAVVNMLTETAAVKFKPEVEVTADTAESLAKRLTESGFEAKRRVSGMGVAENV\
KKWKEMVSKKEDLLVKSRNRVAFAWTLVALCCGSHTSHILHSLGIHIAHGGIWDLLHNSYVKGGLAVGALLGPGRELLF\
DGIKAFGKRSPNMNSLVGLGSMAAFSISLISLVNPELEWDASFFDEPVMLLGFVLLGRSLEERAKLQASTDMNELLSLI\
STQSRLVITSSDNNTPVDSVLSSDSICINVSVDDIRVGDSLLVLPGETFPVDGSVLAGRSVVDESMLTGESLPVFKEEG\
CSVSAGTINWDGPLRIKASSTGSNSTISKIVRMVEDAQGNAAPVQRLADAIAGPFVYTIMSLSAMTFAFWYYVGSHIFP\
DVLLNDIAGPDGDALALSLKLAVDVLVVSCPCALGLATPTAILIGTSLGAKRGYLIRGGDVLERLASIDCVALDKTGTL\
TEGRPVVSGVASLGYEEQEVLKMAAAVEKTATHPIAKAIVNEAESLNLKTPETRGQLTEPGFGTLAEIDGRFVAVGSLE\
WVSDRFLKKNDSSDMVKLESLLDHKLSNTSSTSRYSKTVVYVGREGEGIIGAIAISDCLRQDAEFTVARLQEKGIKTVL\
LSGDREGAVATVAKNVGIKSESTNYSLSPEKKFEFISNLQSSGHRVAMVGDGINDAPSLAQADVGIALKIEAQENAASN\
AASVILVRNKLSHVVDALSLAQATMSKVYQNLAWAIAYNVISIPIAAGVLLPQYDFAMTPSLSGGLMALSSIFVVSNSL\
LLQLHKSETSKNSL'

    def test_fasta1(self):
        # Fasta file with id and description
        fastafile = os.path.join(testdir, 'test1.fa')
        try:
            subprocess.run(['pyTMHMM',
                            '-f',
                            fastafile],
                           check=True)
        except (subprocess.CalledProcessError) as exception:
            print(
                "Error running pyTMHMM -f {0}: {1}".format(fastafile, str(exception)))
        with open(self.annotation) as infile:
            checksum = hashlib.md5(infile.read().encode()).hexdigest()
        self.remove_files()
        self.assertEqual(
            checksum, 'fc4ec46d73f9ab28b5623bded3d6f629', "Correct 128-bit checksum")

    def test_fasta2(self):
        # Fasta file with just id
        fastafile = os.path.join(testdir, 'test2.fa')
        try:
            subprocess.run(['pyTMHMM',
                            '-f',
                            fastafile],
                           check=True)
        except (subprocess.CalledProcessError) as exception:
            print(
                "Error running pyTMHMM -f {0}: {1}".format(fastafile, str(exception)))
        with open(self.annotation) as infile:
            checksum = hashlib.md5(infile.read().encode()).hexdigest()
        self.remove_files()
        self.assertEqual(
            checksum, '43826e2c876aaae1237df4296fd4ad28', "Correct 128-bit checksum")

    def test_annotation(self):
        annotation = pyTMHMM.predict(self.seq, compute_posterior=False)
        self.assertEqual(annotation.count('M'), 64,
                         "Correct number of TM amino acids")
        self.assertEqual(annotation.count('O'), 445,
                         "Correct number of external amino acids")
        self.assertEqual(annotation.count('i'), 365,
                         "Correct number of internal amino acids")
        self.assertEqual(len(re.findall(r'M+', annotation)),
                         3, "Correct number of TMs")

    def test_posterior(self):
        _, posterior = pyTMHMM.predict(self.seq)
        self.assertEqual(len(posterior), 883,
                         "Correct number of aa's in posterior")

    def remove_files(self):
        os.remove(self.annotation)
        os.remove(self.plot)
        os.remove(self.summary)


if __name__ == '__main__':
    unittest.main()



from collections import namedtuple


FastaEntry = namedtuple('FastaEntry', 'id, description, sequence')


def load_posterior_file(fileobj):
    # skip header
    fileobj.readline()
    res = []
    for line in fileobj:
        res.append(map(float, line.split()))
    return zip(*res)


def dump_posterior_file(fileobj, posterior):
    print('inside', 'membrane', 'outside', file=fileobj)
    for i in range(posterior.shape[0]):
        line = '{} {} {}'.format(
            posterior[i, 0], posterior[i, 1], posterior[i, 2])
        print(line, file=fileobj)


def load_fasta_file(fileobj):
    """Load a FASTA-formatted file.

    Returns a list of `(id, description, sequence)` tuples. The `id` and
    `description` is extracted from the header line. The `id` is the part of
    the header line before the first whitespace character and must be unique.
    The `description` is everything coming after the first whitespace character
    and does not need to be unique, and not all FASTA headers have descriptions.
    """
    entries = []
    header = None
    sequence_parts = []

    def append_entry(header, sequence_parts, entries):
        arr = header.split(None, 1)
        if len(arr) == 1:
            arr.append('')
        sequence = ''.join(sequence_parts)
        entries.append(FastaEntry(arr[0], arr[1], sequence))

        header = None
        sequence_parts = []

    for line in fileobj:
        if line.startswith('#'):
            continue
        if line.startswith('>'):
            if header is None:
                header = line[1:].strip()
            else:
                append_entry(header, sequence_parts, entries)
                header = line[1:].strip()
        else:
            sequence_parts.append(line.strip())
    if header is not None:
        append_entry(header, sequence_parts, entries)
    return entries



import argparse
import itertools
import textwrap

from .api import predict
from .api import DEFAULT_MODEL
from .model import parse
from .utils import (
    dump_posterior_file,
    load_posterior_file,
    load_fasta_file,
)

has_matplotlib = True
try:
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
except ImportError:
    has_matplotlib = False


PRETTY_NAMES = {
    'i': 'inside',
    'M': 'transmembrane helix',
    'o': 'outside',
    'O': 'outside'
}


def summarize(path):
    """
    Summarize a path as a list of (start, end, state) triples.
    """
    for state, group in itertools.groupby(enumerate(path), key=lambda x: x[1]):
        group = list(group)
        start = min(group, key=lambda x: x[0])[0]
        end = max(group, key=lambda x: x[0])[0]
        yield start, end, state


def plot(posterior_file, outputfile):
    inside, membrane, outside = load_posterior_file(posterior_file)

    plt.figure(figsize=(16, 8))
    plt.title('Posterior probabilities')
    plt.suptitle('pyTMHMM')
    plt.plot(inside, label='inside', color='blue')
    plt.plot(membrane, label='transmembrane', color='red')
    plt.fill_between(range(len(inside)), membrane, color='red')
    plt.plot(outside, label='outside', color='black')
    plt.legend(frameon=False, bbox_to_anchor=[0.5, 0],
               loc='upper center', ncol=3, borderaxespad=1.5)
    plt.tight_layout(pad=3)
    plt.savefig(outputfile)


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--file', dest='sequence_file',
                        type=argparse.FileType('r'), required=True,
                        help='path to file in fasta format with sequences')
    parser.add_argument('-m', '--model', dest='model_file',
                        default=DEFAULT_MODEL,
                        help='path to the model to use')
    if has_matplotlib:
        parser.add_argument('-p', '--plot', dest='plot_posterior',
                            action='store_true',
                            help='plot posterior probabilies')

    args = parser.parse_args()

    for entry in load_fasta_file(args.sequence_file):
        path, posterior = predict(entry.sequence, args.model_file)

        with open(entry.id + '.summary', 'w') as summary_file:
            for start, end, state in summarize(path):
                print("{} {} {}".format(start, end, PRETTY_NAMES[state]),
                      file=summary_file)

        with open(entry.id + '.annotation', 'w') as ann_file:
            print('>', entry.id, ' ', entry.description, sep='', file=ann_file)
            for line in textwrap.wrap(path, 79):
                print(line, file=ann_file)

        plot_filename = entry.id + '.plot'
        with open(plot_filename, 'w') as plot_file:
            dump_posterior_file(plot_file, posterior)

        if hasattr(args, 'plot_posterior') and args.plot_posterior:
            with open(plot_filename, 'r') as fileobj:
                plot(fileobj, entry.id + '.pdf')



from setuptools import setup, Extension

import numpy

setup(
    name='pyTMHMM',
    version='1.3.2',
    author='Brian Osborne',
    author_email='bosborne@alum.mit.edu',
    description='A transmembrane helix finder.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/bosborne/pyTMHMM/',
    packages=['pyTMHMM'],
    package_data={'pyTMHMM': ['TMHMM2.0.model']},
    zip_safe=False,
    python_version='>=3.5',
    setup_requires=['setuptools>=18.0', 'numpy>=1.9', 'cython'],
    install_requires=['numpy>=1.9'],
    extras_require={
        'plotting':  ['matplotlib'],
    },
    entry_points={
        'console_scripts': ['pyTMHMM=pyTMHMM.cli:cli'],
    },
    ext_modules=[
        Extension(
            'pyTMHMM.hmm',
            sources=['pyTMHMM/hmm.pyx'],
            include_dirs=[numpy.get_include()],
        ),
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: Apache Software License"
    ]
)
